﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("CordeauCrossover", "The merge procedure that is described in Cordeau, J.-F., Gaudioso, M., Laporte, G., Moccia, L. 2006. A memetic heuristic for the generalized quadratic assignment problem. INFORMS Journal on Computing, 18, pp. 433–443.")]
  [StorableClass]
  public class CordeauCrossover : GQAPCrossover,
    IQualitiesAwareGQAPOperator, IWeightsAwareGQAPOperator, IDistancesAwareGQAPOperator, IInstallationCostsAwareGQAPOperator,
    IDemandsAwareGQAPOperator, ICapacitiesAwareGQAPOperator, ITransportationCostsAwareGQAPOperator,
    IOverbookedCapacityPenaltyAwareGQAPOperator {

    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> FlowDistanceQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["FlowDistanceQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> InstallationQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["InstallationQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> OverbookedCapacityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["OverbookedCapacity"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public ILookupParameter<DoubleArray> DemandsParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public ILookupParameter<DoubleArray> CapacitiesParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }
    public IValueLookupParameter<DoubleValue> TransportationCostsParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public IValueLookupParameter<DoubleValue> OverbookedCapacityPenaltyParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["OverbookedCapacityPenalty"]; }
    }

    [StorableConstructor]
    protected CordeauCrossover(bool deserializing) : base(deserializing) { }
    protected CordeauCrossover(CordeauCrossover original, Cloner cloner)
      : base(original, cloner) {
    }
    public CordeauCrossover()
      : base() {
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", GeneralizedQuadraticAssignmentProblem.MaximizationDescription));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", GQAPEvaluator.QualityDescription, 1));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("FlowDistanceQuality", GQAPEvaluator.FlowDistanceQualityDescription, 1));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("InstallationQuality", GQAPEvaluator.InstallationQualityDescription, 1));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("OverbookedCapacity", GQAPEvaluator.OverbookedCapacityDescription, 1));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", GeneralizedQuadraticAssignmentProblem.WeightsDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", GeneralizedQuadraticAssignmentProblem.DistancesDescription));
      Parameters.Add(new LookupParameter<DoubleMatrix>("InstallationCosts", GeneralizedQuadraticAssignmentProblem.InstallationCostsDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", GeneralizedQuadraticAssignmentProblem.DemandsDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", GeneralizedQuadraticAssignmentProblem.CapacitiesDescription));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("TransportationCosts", GeneralizedQuadraticAssignmentProblem.TransportationCostsDescription));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("OverbookedCapacityPenalty", GeneralizedQuadraticAssignmentProblem.OverbookedCapacityPenaltyDescription));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CordeauCrossover(this, cloner);
    }

    public static IntegerVector Apply(IRandom random, BoolValue maximization,
      IntegerVector parent1, DoubleValue quality1,
      IntegerVector parent2, DoubleValue quality2,
      DoubleMatrix weights, DoubleMatrix distances, DoubleMatrix installationCosts,
      DoubleArray demands, DoubleArray capacities,
      DoubleValue transportationCosts, DoubleValue overbookedCapacityPenalty) {
      var mediana = Inizialize(distances);
      int m = capacities.Length;
      int n = demands.Length;
      int i, j, k, ik1, ik2;
      int control;
      bool nofound = false, onefound = false;
      double fbest;
      IntegerVector son = new IntegerVector(parent1.Length), result = null;

      fbest = quality1.Value;
      if (maximization.Value && quality1.Value < quality2.Value
        || !maximization.Value && quality1.Value > quality2.Value) {
        var temp = parent1;
        parent1 = parent2;
        parent2 = temp;
        fbest = quality2.Value;
      }
      var cap = new double[m];
      for (i = 0; i < m; i++) {

        for (j = 0; j < m; j++) cap[j] = 0;

        for (k = 0; k < n; k++) {
          son[k] = -1;
          ik1 = parent1[k];
          ik2 = parent2[k];
          if (distances[i, ik1] < mediana[i]) {
            son[k] = ik1;
            cap[ik1] += demands[k];
          } else if (distances[i, ik2] > mediana[i]) {
            son[k] = ik2;
            cap[ik2] += demands[k];
          }
        }
        k = 0;
        nofound = false;
        while (k < n && !nofound) {
          if (son[k] < 0) {
            control = 0;
            do {
              j = random.Next(m);
              control++;
            }
            while (cap[j] + demands[k] > capacities[j] && control < 3 * m);
            if (cap[j] + demands[k] <= capacities[j]) {
              son[k] = j;
              cap[j] += demands[k];
            } else {
              nofound = true;
            }
          }
          k++;
        }
        if (!nofound) {
          double sonQual = GQAPEvaluator.Evaluate(son, weights, distances, installationCosts, demands, capacities, transportationCosts, overbookedCapacityPenalty);
          if (sonQual < fbest) {
            fbest = sonQual;
            result = (IntegerVector)son.Clone();
            onefound = true;
          }
        }/* else {
          abortedmerge++;
          printf("aborted merge %8d\n", abortedmerge);
        } */
      }
      if (!onefound && !nofound) {
        i = random.Next(m);
        for (j = 0; j < m; j++) {
          cap[j] = 0.0;
        }
        for (k = 0; k < n; k++) {
          son[k] = -1;
          ik1 = parent1[k];
          ik2 = parent2[k];
          if (distances[i, ik1] < mediana[i]) {
            son[k] = ik1;
            cap[ik1] += demands[k];
          } else if (distances[i, ik2] > mediana[i]) {
            son[k] = ik2;
            cap[ik2] += demands[k];
          }
        }
        for (k = 0; k < n; k++) {
          if (son[k] < 0) {
            j = random.Next(m);
            son[k] = j;
            cap[j] += demands[k];
          }
        }
        if (result == null) {
          result = (IntegerVector)son.Clone();
          onefound = true;
        } else {
          double sonQual = GQAPEvaluator.Evaluate(son, weights, distances, installationCosts, demands, capacities, transportationCosts, overbookedCapacityPenalty);
          if (sonQual < fbest) {
            fbest = sonQual;
            result = (IntegerVector)son.Clone();
            onefound = true;
          }
        }
        /*if (tabufix(&son, 0.5 * sqrt(n * m), round(n * m * log10(n)), &tabufix_it)) {
          solution_cost(&son);
          if (son.cost < fbest) {
            fbest = son.cost;
            *sptr = son;
            onefound = TRUE;
            merge_fixed++;
          }*/
      }
      if (result == null) throw new InvalidOperationException("Child could not be created.");
      return result;
    }

    protected override IntegerVector Cross(IRandom random, ItemArray<IntegerVector> parents) {
      if (parents == null) throw new ArgumentNullException("parents");
      if (parents.Length != 2) throw new ArgumentException(Name + " works only with exactly two parents.");

      var qualities = QualityParameter.ActualValue;
      return Apply(random, MaximizationParameter.ActualValue,
        parents[0], qualities[0],
        parents[1], qualities[1],
        WeightsParameter.ActualValue, DistancesParameter.ActualValue, InstallationCostsParameter.ActualValue,
        DemandsParameter.ActualValue, CapacitiesParameter.ActualValue,
        TransportationCostsParameter.ActualValue, OverbookedCapacityPenaltyParameter.ActualValue);
    }

    private static double[] Inizialize(DoubleMatrix distances) {
      int i, j;
      double mdi;
      double delta;
      int ilow, ihigh;
      bool balance;
      int operation;
      int count;
      int m = distances.Rows;

      double[] mediana = new double[distances.Rows];

      for (i = 0; i < m; i++) {
        mdi = 0;
        for (j = 0; j < m; j++) {
          mdi += distances[i, j];
        }
        mediana[i] = mdi / m;
        balance = false;
        delta = 1;
        operation = 0;
        count = 0;
        while (!balance && count < 200) {
          ilow = 0;
          ihigh = 0;
          count++;
          for (j = 0; j < m; j++) {
            if (distances[i, j] < mediana[i]) ilow++;
            else if (distances[i, j] > mediana[i]) ihigh++;
          }
          if (ilow > ((m + 1) / 2)) {
            mediana[i] = mediana[i] - delta;
            if (operation == 1) delta = delta / 2;
            operation = -1;
          } else if (ihigh > ((m + 1) / 2)) {
            mediana[i] = mediana[i] + delta;
            if (operation == -1) delta = delta / 2;
            operation = 1;
          } else balance = true;
        }
      }
      return mediana;
    }
  }
}
